import logging
from typing import List, Optional

from prance import ResolvingParser

from caffoa.content_parser import ContentParser
from caffoa.converter import to_camelcase, parse_type, is_primitive
from caffoa.model import EndPoint, Parameter, Response
from caffoa.object_parser import BaseObjectParser


class PathParser:
    def __init__(self, parser: ResolvingParser):
        self.suffix = ""
        self.prefix = ""
        self.parser = parser
        self.responses = dict()
        self.bodies = dict()
        self.known_types = dict()

    def create_returns(self, model_parser: ResolvingParser):
        for path, options in model_parser.specification['paths'].items():
            for operation, config in options.items():
                operation = operation.lower()
                if operation in ["parameters"]:
                    continue
                if "operationId" not in config:
                    raise Warning(f"operationId is missing for {path}: {operation}")
                operation_id = config["operationId"]
                code = None
                try:
                    responses = list()
                    for code, data in config.get("responses").items():
                        if "$ref" in data:
                            ref = data["$ref"].split('/')[-1]
                            data = model_parser.specification['components']['responses'][ref]
                        response = Response(code)
                        response.content = self.parse_content(data)
                        responses.append(response)
                    self.responses[operation_id] = responses

                except Warning as msg:
                    logging.warning(
                        f"Cannot generate typed responses for {msg} for {operation_id} ({path}): {operation}/{code}")

    def create_bodies(self, model_parser: ResolvingParser):
        one_of_schemas = dict()
        for name, schema in model_parser.specification['components'].get("schemas", dict()).items():
            if "oneOf" in schema:
                one_of_schemas[name] = schema

        for path, options in model_parser.specification['paths'].items():
            for operation, config in options.items():
                operation = operation.lower()
                if operation in ["parameters"]:
                    continue
                if "operationId" not in config:
                    raise Warning(f"operationId is missing for {path}: {operation}")
                operation_id = config["operationId"]
                if "requestBody" in config:
                    try:
                        self.bodies[operation_id] = ContentParser(self.known_types, self.prefix,
                                                                  self.suffix).parse_content_list(config["requestBody"],
                                                                                                  one_of_schemas)
                    except Warning as msg:
                        logging.warning(
                            f"Cannot generate typed requestBody for {msg} for {operation_id} ({path}): {operation}")

    def parse_content(self, config: dict) -> Optional[str]:
        return ContentParser(self.known_types, self.prefix, self.suffix).parse_content_string(config)

    def parse(self) -> List[EndPoint]:
        endpoints = list()
        for path, options in self.parser.specification['paths'].items():
            path = str(path).strip('/')
            base_parameters = list()
            if "parameters" in options:
                base_parameters = self.parse_params(options["parameters"])
            for operation, config in options.items():
                operation = operation.lower()
                if operation in ["parameters"]:
                    continue
                if "operationId" not in config:
                    raise Warning(f"OperationId is missing for {path}: {operation}")
                operation_id = config['operationId']
                operation_name = self.operation_name(operation_id)
                documentation = config['description'].split("\n")
                for response, response_data in config["responses"].items():
                    documentation.append(f"{response} -> {response_data['description']}")

                parameters = base_parameters.copy()
                if "parameters" in config:
                    parameters.extend(self.parse_params(config["parameters"]))
                needs_content = "requestBody" in config
                ep = EndPoint(operation_id, operation_name, path, operation, parameters, documentation, needs_content)
                ep.responses = self.responses.get(operation_id)
                ep.body = self.bodies.get(operation_id)
                endpoints.append(ep)
        return endpoints

    @staticmethod
    def operation_name(operation_id) -> str:
        return to_camelcase(operation_id) + "Async"

    @staticmethod
    def parse_params(params: list) -> list:
        parameters = list()
        for param in params:
            if param['in'].lower() == "path":
                parameters.append(Parameter(param['name'], parse_type(param['schema']), param.get('description')))
        return parameters



from typing import Optional, List, Set


class MemberData:
    def __init__(self, name):
        self.is_list = False
        self.name = name
        self.typename = None
        self.is_date = False
        self.nullable = False
        self.default_value = None
        self.is_required = False
        self.description = None
        self.enums = None
        self.is_basic_type = True

class ModelData:
    def __init__(self, rawname: str, name: str):
        self.rawname = rawname
        self.name = name
        self.description = None

    @staticmethod
    def is_interface() -> bool:
        return False

class ModelObjectData(ModelData):
    def __init__(self, rawname: str, name: str):
        super().__init__(rawname, name)
        self.parent = None
        self.properties = list()
        self.imports = list()
        self.interfaces = list()

class ModelInterfaceData(ModelData):
    def __init__(self, rawname: str, name: str):
        super().__init__(rawname, name)
        self.children = list()
        self.discriminator = None

    @staticmethod
    def is_interface() -> bool:
        return True


class EndPoint:
    def __init__(self, operation: str, name: str, path: str, method: str, parameters: list, documentation: list,
                 needs_content: bool):
        self.operation = operation
        self.name = name
        self.parameters = parameters
        self.method = method.lower()
        self.path = path
        self.documentation = documentation
        self.needs_content = needs_content
        self.responses = None
        self.body = None

    def __str__(self):
        return f"{self.method} {self.path} ({self.parameters}) -> {self.responses}"


class Parameter:
    def __init__(self, name: str, type: str, desc: str):
        self.desc = desc
        self.type = type
        self.name = name

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"{self.type} {self.name}"


class Response:
    def __init__(self, code):
        self.code = int(code)
        self.content = None

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"{self.code}/{self.content}"


class MethodResult:
    def __init__(self, name: str):
        self.base = name
        self.name = name
        self.code = 200
        self.codes = list()
        self.is_simple = True


class BodyConfig:
    def __init__(self, types: List[str]):
        self.types = types
        self.mapping = dict()
        self.discriminator = ""

    def is_selection(self):
        return len(self.types) > 1



from typing import Optional

from caffoa.model import EndPoint


class BodyTypeFilter:
    def __init__(self, config: list):
        self.config = config

    def _match_filter(self, endpoint: EndPoint) -> Optional[str]:
        for item in self.config:
            filter = item.get("filter", dict())
            for method in filter.get("methods", list()):
                if method.lower() == endpoint.method:
                    return item["type"]
            for operation in filter.get("operations", list()):
                if operation == endpoint.operation:
                    return item["type"]
        return None

    def body_type(self, endpoint: EndPoint) -> Optional[str]:
        return self._match_filter(endpoint)

    def additional_imports(self, endpoint: EndPoint) -> list:
        type = self._match_filter(endpoint)
        if type is None:
            return list()
        if type.lower() == "jobject":
            return ["Newtonsoft.Json.Linq"]
        return list()
